# Author: baichen.bai@alibaba-inc.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np


def random_seed(random=np.random.random.__self__):
	"""
		Randomly generate a seed between (0, 2^32 - 1) w.r.t. `random`.
	"""
	return random.randint(0, np.iinfo(np.uint32).max)



# Author: baichen.bai@alibaba-inc.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import argparse
import json
import uuid
from enum import IntEnum, auto
from pathvalidate.argparse import validate_filepath
from iccad_contest.utils.constants import solution_settings
from iccad_contest.utils.path_utils import open_with_abs_path
from iccad_contest.utils.basic_utils import error, assert_error


def type_of_file_path(val):
	validate_filepath(val, platform="auto")
	return val


def type_of_uuid(val):
	def generate_uuid(val):
		_uuid = uuid.UUID(hex=str.lower(val))
		assert _uuid.hex == val, \
			assert_error("error in generating UUID.")
		return _uuid
	return generate_uuid(str.lower(val))


def type_of_join(val):
	return val


def type_of_positive_int(val):
	val = int(val)
	if val <= 0:
		error(
			"a positive integer is expected, but the platform gets {}.".format(val)
		)
	return val


def base_parser():
	parser = argparse.ArgumentParser(add_help=False)
	parser.add_argument(
		"-o",
		"--output-path",
		default=os.path.abspath(
			os.path.join(
				os.getcwd(),
				"outputs"
			)
		),
		type=type_of_file_path,
		help="contest output path specification"
	)
	return parser


def experiment_parser(description):
	parser = argparse.ArgumentParser(
		description=description,
		parents=[base_parser()]
	)
	parser.add_argument(
		"-u",
		"--uuid",
		type=type_of_uuid,
		default=uuid.uuid4().hex,
		help="universally unique identifier (UUID) specification"
	)
	parser.add_argument(
		"-s",
		"--solution-settings",
		type=type_of_join,
		help="solution submission specification"
	)
	parser.add_argument(
		"-q",
		"--num-of-queries",
		default=5,
		type=type_of_positive_int,
		help="the number of queries specification"
	)
	return parser


def args_to_dict(args):
	return dict(vars(args))


def parse_args(parser):
	"""
		arguments parser.
	"""
	return args_to_dict(parser.parse_args())


def load_solution_settings(solution_settings):
	try:
		with open_with_abs_path(
			os.path.join(
				os.path.abspath(solution_settings)
			),
			'r'
		) as f:
			settings = json.load(f)
	except Exception as e:
		settings = {}
	return settings



# Author: baichen.bai@alibaba-inc.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch
import numpy as np
from itertools import combinations
from abc import ABC, abstractmethod
from scipy import stats
from iccad_contest.functions.design_space import parse_contest_design_space
from iccad_contest.functions.dataset import load_contest_dataset
from iccad_contest.utils.constants import dim_of_objective_values
from iccad_contest.utils.basic_utils import assert_error


class Problem(ABC):
    def __init__(self):
        super(Problem, self).__init__()
        self._design_space = None

    @property
    def design_space(self):
        return self._design_space

    @design_space.setter
    def design_space(self, design_space):
        self._design_space = design_space

    @design_space.deleter
    def design_space(self):
        del self._design_space

    @abstractmethod
    def evaluate(self, points):
        raise NotImplementedError()


class PPAT(object):
    def __init__(self, performance, power, area, time_of_vlsi_flow):
        super(PPAT, self).__init__()
        self._performance = performance
        self._power = power
        self._area = area
        self._time_of_vlsi_flow = time_of_vlsi_flow

    @property
    def performance(self):
        return self._performance

    @property
    def power(self):
        return self._power

    @property
    def area(self):
        return self._area

    @property
    def time_of_vlsi_flow(self):
        return self._time_of_vlsi_flow

    def get_objective_values(self):
        return [
            self.performance, self.power, self.area
        ]


class DesignSpaceExplorationProblem(Problem):
    def __init__(self):
        super(DesignSpaceExplorationProblem, self).__init__()
        self.design_space = parse_contest_design_space()
        self.dataset = load_contest_dataset()
        self.dim_of_objective_values = dim_of_objective_values
        self.microarchitecture_embedding_set = self.dataset[
            :, :-(dim_of_objective_values + 1)
        ]
        self.ppa = stats.zscore(self.dataset[:, -4:-1])
        self.performance = self.ppa[:, -3]
        self.power = self.ppa[:, -2]
        self.area = self.ppa[:, -1]
        self.time_of_vlsi_flow = self.dataset[:, -1]
        self.pareto_frontier = self.get_pareto_frontier()
        self.reference_point = self.calc_reference_point()

    def evaluate(self, microarchitecture_embedding):
        idx = [
            all(x) \
                for x in np.equal(
                    self.microarchitecture_embedding_set,
                    microarchitecture_embedding
                )
        ].index(True)
        return PPAT(
            self.performance[idx],
            self.power[idx],
            self.area[idx],
            self.time_of_vlsi_flow[idx]
        )

    def get_pareto_frontier(self):
        return get_pareto_frontier(torch.Tensor(self.ppa))

    def calc_reference_point(self):
        # NOTICE: each vale of the reference point should be the worst.
        return [
            # retrieve the minimal PPA value w.r.t. the Pareto frontier
            # however, in the calculation of hypervolume, we assume minimization,
            # so we negate each value
            -np.min(self.performance),
            -np.min(self.power),
            -np.min(self.area)
        ]


def _get_non_dominated(Y, maximize=True):
    is_efficient = torch.ones(
        *Y.shape[:-1],
        dtype=bool,
        device=Y.device
    )
    for i in range(Y.shape[-2]):
        i_is_efficient = is_efficient[..., i]
        if i_is_efficient.any():
            vals = Y[..., i : i + 1, :]
            if maximize:
                update = (Y > vals).any(dim=-1)
            else:
                update = (Y < vals).any(dim=-1)
            update[..., i] = i_is_efficient.clone()
            is_efficient2 = is_efficient.clone()
            if Y.ndim > 2:
                is_efficient2[~i_is_efficient] = False
            is_efficient[is_efficient2] = update[is_efficient2]
    return is_efficient


def get_non_dominated(Y, deduplicate=True):
    MAX_BYTES = 5e6
    n = Y.shape[-2]
    if n == 0:
        return torch.zeros(
            Y.shape[:-1],
            dtype=torch.bool,
            device=Y.device
        )
    el_size = 64 if Y.dtype == torch.double else 32
    if n > 1000 or \
        n ** 2 * Y.shape[:-2].numel() * el_size / 8 > MAX_BYTES:
        return _get_non_dominated(Y)

    Y1 = Y.unsqueeze(-3)
    Y2 = Y.unsqueeze(-2)
    dominates = (Y1 >= Y2).all(dim=-1) & (Y1 > Y2).any(dim=-1)
    nd_mask = ~(dominates.any(dim=-1))
    if deduplicate:
        indices = (Y1 == Y2).all(dim=-1).long().argmax(dim=-1)
        keep = torch.zeros_like(nd_mask)
        keep.scatter_(dim=-1, index=indices, value=1.0)
        return nd_mask & keep
    return nd_mask


def get_pareto_frontier(objective_values):
    """
        objective_values: <torch.Tensor>
        NOTICE: `get_pareto_frontier` assumes maximization.
    """
    assert isinstance(objective_values, torch.Tensor), \
        assert_error("please convert the input to 'torch.Tensor.")
    # NOTICE: we negate the normalized power and area values
    # to calculate the Pareto frontier since a better design
    # has higher performance and lower power and area values.
    for i in [1, 2]:
        objective_values[:, i] = -objective_values[:, i]
    return objective_values[get_non_dominated(objective_values)]


def get_adrs(reference_pareto_frontier, predict_pareto_frontier):
    """
        reference_pareto_frontier: <torch.Tensor>
        predict_pareto_frontier: <torch.Tensor>
    """
    # calculate average distance to the `reference_pareto_frontier` set
    assert isinstance(reference_pareto_frontier, torch.Tensor) and \
        isinstance(reference_pareto_frontier, torch.Tensor), \
            assert_error("please convert the input to 'torch.Tensor.")
    adrs = 0
    reference_pareto_frontier = reference_pareto_frontier.cpu()
    predict_pareto_frontier = predict_pareto_frontier.cpu()
    for omega in reference_pareto_frontier:
        mini = float('inf')
        for gamma in predict_pareto_frontier:
            mini = min(mini, np.linalg.norm(omega - gamma))
        adrs += mini
    adrs = adrs / len(reference_pareto_frontier)
    return adrs



# Author: baichen.bai@alibaba-inc.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
from collections import OrderedDict
from abc import ABC, abstractmethod
from iccad_contest.functions.dataset import load_contest_design_space
from iccad_contest.utils.basic_utils import warn, error, assert_error


class DesignSpace(ABC):
	def __init__(
		self,
		size,
		dims_of_index_vector,
		dims_of_microarchitecture_embedding
	):
		"""
			size: <int> total size of the design space
			dims_of_index_vector: <int>
				dimension of a microarchitecture embedding index vector
			dims_of_microarchitecture_embedding: <int>
				dimension of a microarchitecture embedding
		"""
		self.size = size
		self.dims_of_index_vector = dims_of_index_vector
		self.dims_of_microarchitecture_embedding = dims_of_microarchitecture_embedding

	@abstractmethod
	def contain(self, design):
		raise NotImplementedError()

	@abstractmethod
	def idx_to_vec(self, idx):
		"""
			transfer from an index to a vector
			idx: <int>
		"""
		raise NotImplementedError()

	@abstractmethod
	def vec_to_idx(self, vec):
		"""
			transfer from a vector to an index
			vec: <list> microarchitecture encoding
		"""
		raise NotImplementedError()


class MicroarchitectureDesignSpace(DesignSpace):
	def __init__(self, descriptions, components_mappings, size):
		"""
			descriptions: <class "collections.OrderedDict">
			Example:
			descriptions = {
				"sub-design-1": {
					"Fetch": [1],
					"Decoder": [1],
					"ISU": [1, 2, 3],
					"IFU": [1, 2, 3],
					...
				}
			}

			components_mappings: <class "collections.OrderedDict">
			Example:
			components_mappings = {
				"Fetch": {
					"description": ["FetchWidth"],
					"1": [4]
				},
				"Decoder": {
					"description": ["DecodeWidth"],
					"1": [1]
				},
				"ISU": {
					"description": [
						"MEM_INST.DispatchWidth", "MEM_INST.IssueWidth"
						"MEM_INST.NumEntries", "INT_INST.DispatchWidth",
						"INT_INST.IssueWidth", "INT_INST.NumEntries",
						"FP_INST.DispatchWidth", "FP_INST.IssueWidth",
						"FP_INST.NumEntries"
					],
					"1": [1, 1, 8, 1, 1, 8, 1, 1, 8],
					"2": [1, 1, 6, 1, 1, 6, 1, 1, 6],
					"3": [1, 1, 10, 1, 1, 12, 1, 1, 12]
				},
				"IFU": {
					"description": ["BranchTag", "FetchBufferEntries", "FetchTargetQueue"]
					"1": [8, 8, 16],
					"2": [6, 6, 14],
					"3": [10, 12, 20]
				},
				...
			}

			size: <int> the size of the entire design space
		"""
		self.descriptions = descriptions
		self.components_mappings = components_mappings
		# construct look-up tables
		# self.designs: <list>
		# Example:
		# 	["sub-design-1", ...]
		self.designs = list(self.descriptions.keys())
		# self.components: <list>
		# Example:
		# 	["Fetch", "Decoder", "ISU", ...]
		self.components = list(self.descriptions[self.designs[0]].keys())
		self.component_offset = self.construct_component_offset()
		self.design_size = self.construct_design_size()
		self.acc_design_size = list(map(
				lambda x, idx: np.sum(x[:idx]),
				[self.design_size for i in range(len(self.design_size))],
				range(1, len(self.design_size) + 1)
			)
		)
		self.component_dims = self.construct_component_dims()
		super(MicroarchitectureDesignSpace, self).__init__(
			size,
			len(self.components),
			sum([
					len(self.components_mappings[c]["description"]) \
						for c in self.components
				]
			)
		)

	def construct_component_offset(self):
		component_offset = []
		for k, v in self.descriptions.items():
			_component_offset = []
			for _k, _v in v.items():
				_component_offset.append(_v[0])
			component_offset.append(_component_offset)
		return component_offset

	def construct_design_size(self):
		design_size = []
		for k, v in self.descriptions.items():
			_design_size = []
			for _k, _v in v.items():
				_design_size.append(len(_v))
			design_size.append(np.prod(_design_size))
		return design_size

	def construct_component_dims(self):
		component_dims = []
		for k, v in self.descriptions.items():
			_component_dims = []
			for _k, _v in v.items():
				_component_dims.append(len(_v))
			component_dims.append(_component_dims)
		return component_dims

	def idx_to_vec(self, idx):
		"""
			idx: <int> the index of a microarchitecture
		"""
		idx -= 1
		assert idx >= 0, assert_error("invalid index.")
		assert idx < self.size, assert_error("index exceeds the size of design space.")
		vec = []
		design = np.where(np.array(self.acc_design_size) > idx)[0][0]
		if design >= 1:
			# NOTICE: subtract the offset
			idx -= self.acc_design_size[design - 1]
		for dim in self.component_dims[design]:
		    vec.append(idx % dim)
		    idx //= dim
		# add the offset
		for i in range(len(vec)):
			vec[i] += self.component_offset[design][i]
		return vec

	def vec_to_idx(self, vec):
		"""
			vec: <list> the list of a microarchitecture encoding
		"""
		def _get_design():
			# a tricky to identify the design
			for k, v in self.descriptions.items():
				if v["Decoder"][0] == vec[1]:
					return v["Decoder"][0] - 1
			error("an invalid index vector.")

		idx = 0
		design = _get_design()
		# subtract the offset
		for i in range(len(vec)):
			vec[i] -= self.component_offset[design][i]
		for j, k in enumerate(vec):
		    idx += int(np.prod(self.component_dims[design][:j])) * k
		if design >= 1:
			# NOTICE: add the offset
			idx += self.acc_design_size[design - 1]
		assert idx >= 0, assert_error("invalid index.")
		assert idx < self.size, assert_error("index exceeds the size of design space.")
		idx += 1
		return idx

	def get_mapping_params(self, vec, idx):
		return self.components_mappings[self.components[idx]][vec[idx]]

	def vec_to_microarchitecture_embedding(self, vec):
		microarchitecture_embedding = []
		for i in range(len(vec)):
			for param in self.get_mapping_params(vec, i):
				microarchitecture_embedding.append(param)
		return microarchitecture_embedding

	def contain(self, microarchitecture_embeddings):
		def _get_design(vec):
			# a tricky to identify the design
			for k, v in self.descriptions.items():
				if v["Decoder"][0] == vec[1]:
					return self.designs[v["Decoder"][0] - 1]
			raise ValueError

		for microarchitecture_embedding in microarchitecture_embeddings:
			vec = microarchitecture_embedding
			try:
				design = _get_design(vec)
			except ValueError:
				raise ValueError
			try:
				offset = 0
				for i in range(len(self.components)):
					exist = False
					l = len(self.components_mappings[self.components[i]]["description"])
					for k, v in self.components_mappings[self.components[i]].items():
						if vec[offset : offset + l] == v:
							exist = True
					if not exist:
						warn(
							"component: {} - {} of the suggest " \
							"microarchitecture embedding: {} is invalid.".format(
								self.components[i],
								vec[offset : offset + l],
								microarchitecture_embedding
							)
						)
						raise ValueError
					offset += l
			except Exception:
				raise ValueError


def parse_microarchitecture_design_space_sheet(microarchitecture_design_space_sheet):
	descriptions = OrderedDict()
	head = microarchitecture_design_space_sheet.columns.tolist()

	# parse design space
	for row in microarchitecture_design_space_sheet.values:
		# extract designs
		descriptions[row[0]] = OrderedDict()
		# extract components
		for col in range(1, len(head) - 1):
			descriptions[row[0]][head[col]] = []
		# extract candidate values
		for col in range(1, len(head) - 1):
			try:
				# multiple candidate values
				for item in list(map(lambda x: int(x), row[col].split(','))):
					descriptions[row[0]][head[col]].append(item)
			except AttributeError:
				# single candidate value
				descriptions[row[0]][head[col]].append(row[col])
	return descriptions


def parse_components_sheet(components_sheet):
	components_mappings = OrderedDict()
	head = components_sheet.columns.tolist()

	# construct look-up tables
	# mappings: <list> [name, width, idx]
	# Example:
	# 	mappings = [("ISU", 10, 0), ("IFU", 4, 10), ("ROB", 2, 14)]
	mappings = []
	for i in range(len(head)):
		if not head[i].startswith("Unnamed"):
			if i == 0:
				name, width, idx = head[i], 1, i
			else:
				mappings.append((name, width, idx))
				name, width, idx = head[i], 1, i
		else:
			width += 1
	mappings.append((name, width, idx))

	for name, width, idx in mappings:
		# extract components
		components_mappings[name] = OrderedDict()
		# extract descriptions
		components_mappings[name]["description"] = []
		for i in range(idx + 1, idx + width):
			components_mappings[name]["description"].append(components_sheet.values[0][i])
		# extract candidate values
		# get number of rows, a trick to test <class "float"> of nan
		nrow = np.where(components_sheet[name].values == \
			components_sheet[name].values)[0][-1]
		for i in range(1, nrow + 1):
			components_mappings[name][int(i)] = \
				list(components_sheet.values[i][idx + 1: idx + width])
	return components_mappings


def parse_contest_design_space():
	"""
		configs: <dict>
	"""
	microarchitecture_design_space_sheet, components_sheet = load_contest_design_space()
	descriptions = parse_microarchitecture_design_space_sheet(
		microarchitecture_design_space_sheet
	)
	components_mappings = parse_components_sheet(
		components_sheet
	)

	return MicroarchitectureDesignSpace(
		descriptions,
		components_mappings,
		int(microarchitecture_design_space_sheet.values[0][-1])
	)



# Author: baichen.bai@alibaba-inc.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
from iccad_contest.utils.path_utils import path_join_for_read


# solution configuration specification
solution_settings = "configs.json"


# contest dataset root specification
root_name_of_contest_dataset = "contest-dataset"


# contest design space specification
name_of_contest_design_space = "design-space.xlsx"
contest_design_space = path_join_for_read(
	os.path.dirname(__file__),
	os.path.pardir,
	root_name_of_contest_dataset,
	name_of_contest_design_space
)


# contest dataset specification
name_of_contest_dataset = "contest.csv"
contest_dataset = path_join_for_read(
	os.path.dirname(__file__),
	os.path.pardir,
	root_name_of_contest_dataset,
	name_of_contest_dataset
)


# output file name specification
format_of_exp_output_root = "iccad-contest-%Y%m%d-%H%M%S"
derived_root = "derived"
log_root = "log"
evaluation_root = "evaluation"
time_root = "time"
explored_microarchitecture_embedding_root = "explored-microarchitecture-embedding"
list_of_exp_output_sub_root = [
	derived_root,
	log_root,
	evaluation_root,
	time_root,
	explored_microarchitecture_embedding_root
]
suffix_of_log = ".log"


# summary report specification
summary_root = "summary"
suffix_of_summary = ".rpt"


# objectives specification
# performance, power, and area
dim_of_objective_values = 3



# Author: baichen.bai@alibaba-inc.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
    abstract base class for the optimizers.
    it creates a common API across all packages.
"""


from abc import ABC, abstractmethod
from importlib_metadata import version


class AbstractOptimizer(ABC):
    """
        abstract base class for the optimizers.
        it creates a common API across all packages.
    """

    # every implementation package needs to specify `primary_import`,
    # e.g.,
    # primary_import = opentuner
    primary_import = None

    def __init__(self, design_space, **kwargs):
        """
            build a wrapper class for an optimizer.

            parameters
            ----------
            design_space: <class "MicroarchitectureDesignSpace">
        """
        self.design_space = design_space
        # early stopping criterion
        self.early_stopping = False

    @classmethod
    def get_version(cls):
        """
            get the version for the optimizer.

            returns
            -------
            version_str: <str>
                the version number of an optimizer.
                usually, it is equivalent to `package.__version__`.
        """
        assert (cls.primary_import is None) or isinstance(cls.primary_import, str)
        # verstion "x.x.x" is used as a default version
        version_str = "x.x.x" \
            if cls.primary_import is None else version(cls.primary_import)
        return version_str

    @abstractmethod
    def suggest(self):
        """
            get a suggestion from the optimizer.

            returns
            -------
            next_guess: <list> of <list>
                list of suggestions.
                each suggestion is a microarchitecture embedding.
        """
        pass

    @abstractmethod
    def observe(self, x, y):
        """
            send an observation of a suggestion back to the optimizer.

            parameters
            ----------
            x: <list> of <list>
                the output of `suggest`.
            y: <list> of <list>
                corresponding values where each `x` is mapped to.
        """
        pass



# Author: baichen.bai@alibaba-inc.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
from pathvalidate.argparse import validate_filename, validate_filepath
from iccad_contest.utils.basic_utils import mkdir, warn, assert_error


def validate_file_path(path):
	validate_filepath(path, platform="auto")


def validate_file_name(file):
	validate_filename(file, platform="universal")


def open_with_abs_path(path, mode):
	assert os.path.isabs(path), \
		assert_error(
			"an absolute path is expected, but the platform gets {}".format(path)
		)
	return open(path, mode)


def validate_abs_path(path, strict=False):
	path = os.path.abspath(os.path.expanduser(path))
	if strict:
		assert os.path.isdir(path), assert_error("{} does not exist.".format(path))
	return path


def _path_join(*args):
	assert len(args) > 1
	path, file = args[:-1], args[-1]
	path = validate_abs_path(os.path.join(*path), strict=False)
	assert os.path.basename(file) == file, \
		assert_error("a file name is expected, but the platform gets {}".format(file))
	return os.path.join(path, file)


def _strict_path_join(*args):
	assert len(args) > 1
	path, file = args[:-1], args[-1]
	path = validate_abs_path(os.path.join(*path), strict=True)
	assert os.path.basename(file) == file, \
		assert_error("a file name is expected, but the platform gets {}".format(file))
	return os.path.join(path, file)


def path_join_for_read(*args):
	file = _strict_path_join(*args)
	assert os.path.isfile(file), \
		assert_error("a file is expected, but the platform gets {}".format(file))
	return file


def path_join_for_write(*args):
	file = _strict_path_join(*args)
	if os.path.isfile(file):
		warn("{} already exists.".format(file))
	return file

def path_join_for_write_with_create(*args):
	file = _path_join(*args)
	if not os.path.isfile(file):
		mkdir(os.path.dirname(file))
	return file



# Author: baichen.bai@alibaba-inc.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch
from iccad_contest.utils.basic_utils import assert_error


class HyperVolume:
    """
        we assume minimization.
        reference:
            <1> C. M. Fonseca, L. Paquete, and M. Lopez-Ibanez. An improved dimension-sweep
            algorithm for the hypervolume indicator. In IEEE Congress on Evolutionary
            Computation, pages 1157-1163, Vancouver, Canada, July 2006.
            <2> Simon Wessing@TU Dortmund University
    """

    def __init__(self, reference_point):
        """
            reference_point: <list>
        """
        self.reference_point = reference_point
        self.list = []

    def compute(self, predict_point):
        """
            precit_point: <torch.Tensor>
            compute the hypervolume that is dominated by a non-dominated `predict_point`.
            `predict_point` and `reference_point` are translated, so
            that the reference point is [0, ..., 0], before the HV computation
        """

        def weakly_dominates(point, other):
            for i in range(len(point)):
                if point[i] > other[i]:
                    return False
            return True

        assert isinstance(predict_point, torch.Tensor), \
            assert_error("please convert the input to 'torch.Tensor.")

        predict_point = (-predict_point).tolist()
        relevant_points = []
        reference_point = self.reference_point
        dimensions = len(reference_point)
        for point in predict_point:
            # only consider points that dominate the reference point
            if weakly_dominates(point, reference_point):
                relevant_points.append(point)
        if any(reference_point):
            # shift points so that `reference_point` == [0, ..., 0]
            # the `reference_point` have no need to be explicitly used
            # in the HV computation
            for j in range(len(relevant_points)):
                relevant_points[j] = [relevant_points[j][i] - reference_point[i] for i in range(dimensions)]
        self.pre_process(relevant_points)
        bounds = [-1.0e308] * dimensions
        hyper_volume = self.calc_hv(dimensions - 1, len(relevant_points), bounds)
        return hyper_volume

    def calc_hv(self, idx, length, bounds):
        """
            in contrast to the paper, the implementation assumes that `reference_point` = [0, ..., 0], 
            which can reduce a few operations.
        """
        hyper_volume = 0.0
        sentinel = self.list.sentinel
        if length == 0:
            return hyper_volume
        elif idx == 0:
            # NOTICE: only one dimension
            return -sentinel.next[0].cargo[0]
        elif idx == 1:
            # NOTICE: two dimensions
            q = sentinel.next[1]
            h = q.cargo[0]
            p = q.next[1]
            while p is not sentinel:
                p_cargo = p.cargo
                hyper_volume += h * (q.cargo[1] - p_cargo[1])
                if p_cargo[0] < h:
                    h = p_cargo[0]
                q = p
                p = q.next[1]
            hyper_volume += h * q.cargo[1]
            return hyper_volume
        else:
            remove = self.list.remove
            reinsert = self.list.reinsert
            calc_hv = self.calc_hv
            p = sentinel
            q = p.prev[idx]
            while q.cargo is not None:
                if q.ignore < idx:
                    q.ignore = 0
                q = q.prev[idx]
            q = p.prev[idx]
            while length > 1 and (
                    q.cargo[idx] > bounds[idx] or q.prev[idx].cargo[idx] >= bounds[idx]):
                p = q
                remove(p, idx, bounds)
                q = p.prev[idx]
                length -= 1
            q_area = q.area
            q_cargo = q.cargo
            q_prev_index = q.prev[idx]
            if length > 1:
                hyper_volume = q_prev_index.volume[idx] + q_prev_index.area[idx] * (
                    q_cargo[idx] - q_prev_index.cargo[idx]
                )
            else:
                q_area[0] = 1
                q_area[1:idx + 1] = [q_area[i] * -q_cargo[i] for i in range(idx)]
            q.volume[idx] = hyper_volume
            if q.ignore >= idx:
                q_area[idx] = q_prev_index.area[idx]
            else:
                q_area[idx] = calc_hv(idx - 1, length, bounds)
                if q_area[idx] <= q_prev_index.area[idx]:
                    q.ignore = idx
            while p is not sentinel:
                p_cargo_index = p.cargo[idx]
                hyper_volume += q.area[idx] * (p_cargo_index - q.cargo[idx])
                bounds[idx] = p_cargo_index
                reinsert(p, idx, bounds)
                length += 1
                q = p
                p = p.next[idx]
                q.volume[idx] = hyper_volume
                if q.ignore >= idx:
                    q.area[idx] = q.prev[idx].area[idx]
                else:
                    q.area[idx] = calc_hv(idx - 1, length, bounds)
                    if q.area[idx] <= q.prev[idx].area[idx]:
                        q.ignore = idx
            hyper_volume -= q.area[idx] * q.cargo[idx]
            return hyper_volume

    def pre_process(self, predict_point):
        """
            sets up the multi-linked list data structure.
        """
        dimensions = len(self.reference_point)
        node_list = MultiList(dimensions)
        nodes = [MultiList.Node(dimensions, point) for point in predict_point]
        for i in range(dimensions):
            self.sort_by_dimension(nodes, i)
            node_list.extend(nodes, i)
        self.list = node_list

    def sort_by_dimension(self, nodes, i):
        """
            sorts the list of nodes by the i-th value of the contained points.
        """
        # build a list of tuples: (point[i], node)
        decorated = [(node.cargo[i], index, node) for index, node in enumerate(nodes)]
        # sort by this value
        decorated.sort()
        # write back to original list
        nodes[:] = [node for (_, _, node) in decorated]


class MultiList:
    """
        multi-linked list consists of several doubly linked lists that share common nodes.
        every node has multiple predecessors and successors, one in every list.
    """

    class Node:

        def __init__(self, number_of_lists, cargo=None):
            self.cargo = cargo
            self.next = [None] * number_of_lists
            self.prev = [None] * number_of_lists
            self.ignore = 0
            self.area = [0.0] * number_of_lists
            self.volume = [0.0] * number_of_lists

        def __str__(self):
            return str(self.cargo)

    def __init__(self, number_of_lists):
        """
            builds `number_of_lists` doubly linked lists.
        """
        self.number_of_lists = number_of_lists
        self.sentinel = MultiList.Node(number_of_lists)
        self.sentinel.next = [self.sentinel] * number_of_lists
        self.sentinel.prev = [self.sentinel] * number_of_lists

    def __str__(self):
        strings = []
        for i in range(self.number_of_lists):
            currentList = []
            node = self.sentinel.next[i]
            while node != self.sentinel:
                currentList.append(str(node))
                node = node.next[i]
            strings.append(str(currentList))
        ret = ""
        for string in strings:
            ret += string + '\n'
        return ret

    def __len__(self):
        """
            get the number of lists that are included in the `MultiList`
        """
        return self.number_of_lists

    def get_length(self, i):
        """
            get the length of the i-th list.
        """
        length = 0
        sentinel = self.sentinel
        node = sentinel.next[i]
        while node != sentinel:
            length += 1
            node = node.next[i]
        return length

    def append(self, node, index):
        """
            appends a node to the end of the list at the given index.
        """
        last_but_one = self.sentinel.prev[index]
        node.next[index] = self.sentinel
        node.prev[index] = last_but_one
        # set the last element as the new one
        self.sentinel.prev[index] = node
        last_but_one.next[index] = node

    def extend(self, nodes, index):
        """
            extends the list at the given index with the nodes.
        """
        sentinel = self.sentinel
        for node in nodes:
            last_but_one = sentinel.prev[index]
            node.next[index] = sentinel
            node.prev[index] = last_but_one
            # set the last element as the new one
            sentinel.prev[index] = node
            last_but_one.next[index] = node

    def remove(self, node, index, bounds):
        """
            removes and returns `node` from all lists in [0: `index`].
        """
        for i in range(index):
            prev = node.prev[i]
            succ = node.next[i]
            prev.next[i] = succ
            succ.prev[i] = prev
            if bounds[i] > node.cargo[i]:
                bounds[i] = node.cargo[i]
        return node

    def reinsert(self, node, index, bounds):
        """
            inserts `node` at the position it has in all lists in [0: `index`]
            before it was removed.
            the method assumes that the next and previous
            nodes of the node that is reinserted are in the list.
        """
        for i in range(index):
            node.prev[i].next[i] = node
            node.next[i].prev[i] = node
            if bounds[i] > node.cargo[i]:
                bounds[i] = node.cargo[i]



# Author: baichen.bai@alibaba-inc.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys, os
sys.path.insert(
    0,
    os.path.join(os.path.dirname(__file__), "functions")
)
sys.path.insert(
    0,
    os.path.join(os.path.dirname(__file__), "utils")
)
import random
import torch
import numpy as np
from time import time
from collections import OrderedDict
from iccad_contest.functions.problem import PPAT, DesignSpaceExplorationProblem, get_pareto_frontier
from iccad_contest.functions.hypervolume import HyperVolume
from iccad_contest.__init__ import __version__
from iccad_contest.utils.constants import summary_root
from iccad_contest.utils.serialize import ContestSerializer
from iccad_contest.utils.arguments import experiment_parser, parse_args, \
	load_solution_settings
from iccad_contest.utils.numpy_utils import random_seed
from iccad_contest.utils.basic_utils import create_logger, error, assert_error


logger = None


def initialize_logger(output_path, uuid):
	global logger
	log_file = ContestSerializer.logging_path(output_path, uuid)
	logger = create_logger(log_file)
	return logger


def create_problem():
	return DesignSpaceExplorationProblem()


def set_random_seed(seed):
	random_seed_generator = random.Random(int(seed, 16))
	random.seed(random_seed(random_seed_generator))
	np.random.seed(random_seed(random_seed_generator))
	logger.info("[INFO] use UUID {} to reproduce the experiment results.".format(seed))


def do_suggest(
	iteration,
	design_space,
	optimizer,
	suggest_time
):
	t = time()
	try:
		next_microarchitecture_embeddings = optimizer.suggest()
		if next_microarchitecture_embeddings is None:
			raise Exception
	except Exception as e:
		logger.warning(
			"[WARN]: failed in 'suggest' of the optimizer. " \
			"the error message: {}" \
			"\nfall back to random search.".format(e)
		)
		random_index = random.sample(
			range(design_space.size),
			k=1
		)
		next_microarchitecture_embeddings = [
			design_space.vec_to_microarchitecture_embedding(
				design_space.idx_to_vec(_random_index)
			) for _random_index in random_index
		]
	suggest_time.append(time() - t)
	logger.info(
		"[INFO]: {}-th query, " \
		"next microarchitecture embeddings: {} " \
		"'suggest' time: {}.".format(
			iteration + 1,
			str(next_microarchitecture_embeddings),
			suggest_time[iteration]
		)
	)

	try:
		design_space.contain(next_microarchitecture_embeddings)
	except Exception:
		raise ValueError(
			"[ERROR]: suggestions provided by the optimizer are out of the design space."
		)

	return next_microarchitecture_embeddings


def do_evaluate(
	iteration,
	problem,
	next_microarchitecture_embeddings,
	explored_microarchitecture_embeddings,
	objective_values,
	evaluation_time
):
	for i, next_microarchitecture_embedding \
		in enumerate(next_microarchitecture_embeddings):
		t = time()
		try:
			ppat = problem.evaluate(next_microarchitecture_embedding)
		except Exception as e:
			logger.warning(
				"[WARN]: failed in evaluating microarchitecture embeddings. " \
				"set the evaluation of the limit."
			)
			ppat = PPAT(0, float("inf"), float("inf"), float("inf"))
		evaluation_time.append(time() - t + ppat.time_of_vlsi_flow)
		explored_microarchitecture_embeddings.append(next_microarchitecture_embedding)
		objective_values.append(ppat.get_objective_values())
		logger.info(
			"[INFO]: evaluate {}-th candidate in a batch, " \
			"microarchitecture embedding: {}, " \
			"norm. performance: {}, " \
			"norm. power: {}, " \
			"norm. area: {}, " \
			"time of the VLSI flow: {}.".format(
				i + 1,
				next_microarchitecture_embedding,
				ppat.performance,
				ppat.power,
				ppat.area,
				ppat.time_of_vlsi_flow
			)
		)


def do_observe(
	iteration,
	optimizer,
	next_microarchitecture_embeddings,
	objective_values,
	observe_time
):
	t = time()
	try:
		optimizer.observe(
			next_microarchitecture_embeddings,
			objective_values[-len(next_microarchitecture_embeddings):]
		)
	except Exception as e:
		logger.warning(
			"[WARN]: failed in 'observe' of the optimizer. " \
			"the error message: {}" \
			"\nignore observations.".format(e)
		)
		logger.exception(e, exc_info=True)
	observe_time.append(time() - t)


def do_experiment_impl(
	problem,
	optimizer,
	num_of_queries
):
	suggest_time = []
	evaluation_time = []
	observe_time = []
	explored_microarchitecture_embeddings = []
	objective_values = []
	hv = HyperVolume(problem.reference_point)
	golden_hypervolume = hv.compute(problem.pareto_frontier)

	for i in range(num_of_queries):
		next_microarchitecture_embeddings = do_suggest(
			i,
			problem.design_space,
			optimizer,
			suggest_time
		)
		do_evaluate(
			i,
			problem,
			next_microarchitecture_embeddings,
			explored_microarchitecture_embeddings,
			objective_values,
			evaluation_time
		)
		do_observe(
			i,
			optimizer,
			next_microarchitecture_embeddings,
			objective_values,
			observe_time
		)
		curr_hypervolume = hv.compute(
			get_pareto_frontier(
				torch.Tensor(objective_values)
			)
		)
		logger.info(
			"[INFO]: summary for {}-th query, " \
			"current Pareto hypervolume: {}, " \
			"Pareto hypervolume difference: {}, " \
			"cost: {}.".format(
				i + 1,
				curr_hypervolume,
				golden_hypervolume - curr_hypervolume,
				np.sum([
						np.sum(suggest_time),
						np.sum(evaluation_time),
						np.sum(observe_time)
					]
				)
			)
		)
		if optimizer.early_stopping:
			break

	curr_hypervolume = hv.compute(
		get_pareto_frontier(
			torch.Tensor(objective_values)
		)
	)
	logger.info(
		"[INFO]: summary for the solution, " \
		"the best Pareto hypervolume: {}, " \
		"the best Pareto hypervolume difference: {} " \
		"cost: {}.".format(
			curr_hypervolume,
			golden_hypervolume - curr_hypervolume,
			np.sum([
					np.sum(suggest_time),
					np.sum(evaluation_time),
					np.sum(observe_time)
				]
			)
		)
	)

	return (explored_microarchitecture_embeddings, objective_values, golden_hypervolume - curr_hypervolume), \
		(suggest_time, evaluation_time, observe_time)


def experiment_impl(args, solution, solution_settings):
	problem = create_problem()
	optimizer = solution(problem.design_space, **solution_settings)
	return do_experiment_impl(
		problem,
		optimizer,
		args["num_of_queries"]
	)


def parse_experiment_args():
	args = parse_args(
		experiment_parser(
		"ICCAD'22 Contest Platform - solutions evaluation"
		)
	)
	return args


def save_experiment(args, solution_settings, explored_result, timing):
	save_dict = OrderedDict()
	_args = args.copy()
	_args["uuid"] = _args["uuid"].hex
	save_dict["command"] = _args
	save_dict["solution-settings"] = solution_settings
	save_dict["explored-microarchitecture-embedding"] = explored_result[0]
	save_dict["objective-values"] = explored_result[1]
	save_dict["Pareto hypervolume difference"] = explored_result[2]
	save_dict["suggest-time"] = timing[0]
	save_dict["evaluate-time"] = timing[1]
	save_dict["observe-time"] = timing[2]
	save_dict["total-time"] = np.sum([
			np.sum(timing[0]),
			np.sum(timing[1]),
			np.sum(timing[2])
		]
	)
	save_path = ContestSerializer.save(args["output_path"], args["uuid"], save_dict)
	logger.info(
		"[INFO]: save the experiment result. " \
		"please refer them to {}.".format(
			save_path
		)
	)


def experiment(solution):
	"""
		the main function for ICCAD'22 contests.
	"""
	args = parse_experiment_args()
	logger = initialize_logger(args["output_path"], args["uuid"])
	logger.info("[INFO]: Prolem C @ ICCAD'22 contest platform version: {}".format(__version__))
	solution_settings = load_solution_settings(args["solution_settings"])
	if solution_settings == {}:
		logger.info("[INFO]: no --solution-settings is specified.")
	else:
		logger.info("[INFO]: load solution settings: {}".format(solution_settings))
	set_random_seed(args["uuid"].hex)
	explored_result, timing = experiment_impl(args, solution, solution_settings)
	save_experiment(args, solution_settings, explored_result, timing)



# Author: baichen.bai@alibaba-inc.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import logging


def info(msg):
    """
        msg: <str>
    """
    print("[INFO]: {}".format(msg))


def test(msg):
    """
        msg: <str>
    """
    print("[TEST]: {}".format(msg))


def warn(msg):
    """
        msg: <str>
    """
    print("[WARN]: {}".format(msg))


def error(msg):
    """
        msg: <str>
    """
    print("[ERROR]: {}".format(msg))
    exit(1)


def assert_error(msg):
	"""
		msg: <str>
	"""
	return "[ERROR]: {}".format(msg)


def if_exist(path, strict=False):
    try:
        if os.path.exists(path):
            return True
        else:
            raise FileNotFoundError(path)
    except FileNotFoundError as e:
        warn(e)
        if not strict:
            return False
        else:
            exit(1)


def mkdir(path):
    if not if_exist(path):
        info("create directory: %s" % path)
        os.makedirs(path, exist_ok=True)


def remove(path):
    if if_exist(path):
        if os.path.isfile(path):
            os.remove(path)
            info("remove %s" % path)
        elif os.path.isdir(path):
            if not os.listdir(path):
                # empty directory
                os.rmdir(path)
            else:
                shutil.rmtree(path)
            info("remove %s" % path)


def create_logger(log_file):
    logging.basicConfig(filename=log_file, format="%(asctime)-15s %(message)s")
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    console = logging.StreamHandler()
    logging.getLogger('').addHandler(console)
    return logger



# Author: baichen.bai@alibaba-inc.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# A serialization abstraction layer (SAL) to save and load experimental results.\
# All IO experimental results should go through this module.
# This makes changing the backend (between different databases) transparent to the benchmark codes.


import json
import os
import uuid
from abc import ABC, abstractmethod
from datetime import datetime
from tempfile import mkdtemp
import xarray as xr
from iccad_contest.utils.constants import format_of_exp_output_root, list_of_exp_output_sub_root, \
	suffix_of_log, log_root, summary_root, suffix_of_summary
from iccad_contest.utils.path_utils import validate_file_path, validate_file_name, \
	path_join_for_write_with_create


class Serializer(ABC):
	@staticmethod
	@abstractmethod
	def init_contest_output_data(output_path):
		raise NotImplementedError

	@staticmethod
	@abstractmethod
	def logging_path(output_path, uuid):
		raise NotImplementedError

	@staticmethod
	@abstractmethod
	def save(data):
		raise NotImplementedError

	@staticmethod
	@abstractmethod
	def load(path):
		raise NotImplementedError


class ContestSerializer(Serializer):

	def init_contest_output_data(output_path):
		validate_file_path(output_path)

		prefix_of_exp_output_root = datetime.utcnow().strftime(format_of_exp_output_root)
		exp_output_root = mkdtemp(prefix=prefix_of_exp_output_root, dir=output_path)
		for exp_output_sub_root in list_of_exp_output_sub_root:
			mkdir(exp_output_sub_root)
		return os.path.basename(exp_output_root)

	def logging_path(output_path, uuid):
		validate_file_path(output_path)
		return path_join_for_write_with_create(output_path, log_root, uuid.hex + suffix_of_log)

	def load(path):
		validate_file_path(path)
		with open(path, 'r') as f:
			data = json.load(f)
		return data

	def save(output_path, uuid, data):
		validate_file_path(output_path)
		with open(
			path_join_for_write_with_create(output_path, summary_root, uuid.hex + suffix_of_summary),
			'w',
			encoding='utf-8'
		) as f:
			json.dump(data, f, indent=4)
		return path_join_for_write_with_create(output_path, summary_root, uuid.hex + suffix_of_summary)



# Author: baichen.bai@alibaba-inc.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
from pathlib import Path
from setuptools import find_packages, setup


PLATFORM_NAME = "iccad-contest"


# load the long description from `README.md`
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


def read_requirements(name):
    with open(os.path.join("requirements", "{}.in".format(name))) as f:
        requirements = f.read().strip()
    requirements = requirements.replace("==", ">=").splitlines()
    return [p for p in requirements if p[0].isalnum()]


# derive install requirements from base.in requirements.
requirements = read_requirements("base")
opt_requirements = read_requirements("optimizers")
ipynb_requirements = read_requirements("ipynb")


setup(
    name=PLATFORM_NAME,
    version="0.0.7",
    packages=find_packages(),
    package_data={
        "iccad_contest.contest-dataset": [
            "*.csv",
            "*.xlsx"
        ]
    },
    url="https://code.aone.alibaba-inc.com/ctl-research/iccad-contest-platform",
    author="Chen BAI",
    author_email=("baichen.bai@alibaba-inc.com"),
    license="Apache v2",
    description="ICCAD'22 Contest @ Microarchitecture Design Space Exploration",
    install_requires=requirements,
    extras_require={
        "optimizers": opt_requirements,
        "notebooks": ipynb_requirements
    },
    long_description=long_description,
    long_description_content_type="text/markdown",
    platforms=["any"]
)



# Author: baichen.bai@alibaba-inc.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import numpy as np
import pandas as pd
from iccad_contest.utils.constants import contest_design_space, contest_dataset
from iccad_contest.utils.basic_utils import if_exist, info, warn, assert_error


def load_txt(path, fmt=int):
    if if_exist(path):
        info("loading from %s" % path)
        return np.loadtxt(path, dtype=fmt)
    else:
        warn("cannot load %s" % path)


def load_excel(path, sheet_name=0):
    """
        path: <str>
        sheet_name: <int> | <str>
    """
    if_exist(path, strict=True)
    data = pd.read_excel(path, sheet_name=sheet_name)
    info("read the sheet {} of excel from {}".format(sheet_name, path))
    return data


def load_contest_dataset():
	def validate_contest_dataset(dataset):
		assert dataset.ndim == 2, \
			assert_error(
				"{} dimensions are expected, but the platform gets {}.".format(dataset.ndim)
			)
		assert np.all(np.isfinite(dataset)), \
			assert_error(
				"dataset contains infinity."
			)
	dataset = load_txt(contest_dataset, fmt=float)
	validate_contest_dataset(dataset)
	return dataset


def load_contest_design_space():
	microarchitecture_design_space_sheet = load_excel(
		contest_design_space,
		sheet_name="Microarchitecture Design Space"
	)
	components_sheet = load_excel(
		contest_design_space,
		sheet_name="Components"
	)
	return microarchitecture_design_space_sheet, components_sheet
